/***************************************************************************
                          movingobject.cpp  -  description
                             -------------------
    begin                : Tue Oct 23 2001
    copyright            : (C) 2001, 2002, 2004 by Thomas Friedrichsmeier
    email                : Thomas.Friedrichsmeier@ruhr-uni-bochum.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "defines.h"

#include "movingobject.h"

#include <qdom.h>
#include <qstring.h>
#include <qlist.h>
#include <qbitmap.h>
#include <qpainter.h>

#include "taxipilot.h"
#include "level.h"
#include "cdp.h"
#include "platform.h"
#include "teleporter.h"
#include "conditional.h"
#include "state.h"
#include "level.h"
#include "event.h"

MovingObject::MovingObject(const QDomElement *description):AnimatedSprite()
{
	advance_timer = animation_timer = -1;
	setAnimated (false);

	read_object (description);

	current_segment = 0;
	step = 0;
	setAnimation (path[0].anim);		// so other classes can already access width and height and stuff
	if (cdp->for_real ()) {
		advance_timer = tm->startTimer (this, OBJECTS_AP);         // sets everything off
	}
	last_visible = false; 		// attached-objects (platforms/teleporters...) will assume invisiblity, until told otherwise.
}

MovingObject::~MovingObject()
{
	hide ();
	tm->killTimers (this);
	
	QList <Animation> deleted_animations;		// to prevent multiple deletion of shared animations
	deleted_animations.setAutoDelete (false);

	QList <Action> deleted_actions;
	deleted_actions.setAutoDelete (false);

	for (int i=0; i < segments; i++) {
		if (deleted_animations.find (path[i].anim) == -1) {
			deleted_animations.append (path[i].anim);
			delete path[i].anim;			
		}

		if (path[i].action) {
			if (deleted_actions.find (path[i].action) == -1) {
				deleted_actions.append (path[i].action);
				Action *action_a;
				Action *action_b;
				action_a = path[i].action;
				do {                             			// descend into chain of child-actions
					action_b = action_a->next;
					delete action_a->conditional;
					delete action_a;
					action_a = action_b;
				} while (action_b);
			}
		}
	}
	deleted_animations.clear ();

	if (segments) {
		delete [] path;
	}
}

/** This function reads a moving-object-description ( specs for the animations ) from a
QDomNode (of a level-description) read_object is only called by the constructor. */
void MovingObject::read_object(const QDomElement *description)
{
	Animation *globalanim = 0;
	int sound_opts = ReadSingleSound | ReadContinuous | ReadRepetitious | ReadInterruptable;

	// used for debug-message if anim missing in individual segments
	bool have_global_anim = false;

	int global_z;
	QDomElement e;
	QDomNodeList list;
	QString dummy;
	bool looping;
	int max_segment;

	if (description->hasAttribute ("ident")) {
		has_ident = true;
		ident = cdp->get_string_attribute ("ident", *description, bad_string_value, Serious);
	} else {
		has_ident = false;
	}

	teleporter = center = platform = false;
	harmful_if_visible = true;

	QString classification = cdp->get_multi_choice_attribute ( "class", *description, "harmful;background;teleporter;harmful center;harmless center;platform", "harmful", Warn );
	if (classification == "harmful") {

	} else if (classification == "background") {
		harmful_if_visible = false;
	} else if (classification == "platform") {
		platform = true;
		harmful_if_visible = true;		// you can crash into platforms
	} else {
		if (classification == "teleporter") {
			teleporter = true;
			harmful_if_visible = false;
		} else if (classification == "harmful center") {
			center = true;
		} else if (classification == "harmless center") {
			harmful_if_visible = false;
			center = true;
		}
		if (!has_ident) {
			cdp->debug_msg (i18n ("Object classified as '%1' teleporter but no ident given").arg(classification), Serious, description);
		}
	}
	global_z = cdp->get_int_attribute ( "z", *description, 0, 300, 150, Info );
	if ((looping = cdp->get_bool_attribute ( "looping", *description, true, Warn ))) {
		sound_opts = DefaultSingleSound | ReadContinuous | ReadRepetitious | ReadInterruptable;
	}

	// does the object have a global-animation of some kind?
	if (description->hasAttribute ("text") || description->hasAttribute ("file_pattern")) {
		globalanim = new Animation (description, ReadSound | ReadAnimContinuous | ReadAnimRepetitions | ReadAnimOscillating | DefaultAnimContinuous,
										sound_opts);
		have_global_anim = true;
	}

	list = cdp->get_node_list ( "path", *description, 1, -1, 1, Serious );	// now get the path-segments

	segments = 0;

	float previous_vx, previous_vy;

	previous_vx = cdp->get_float_attribute ( "initial_vx", *description, 0, 1000, 0, Info ) / ( (int) (1000 / OBJECTS_AP));
	previous_vy = cdp->get_float_attribute ( "initial_vy", *description, 0, 1000, 0, Info ) / ( (int) (1000 / OBJECTS_AP));

	if ( looping ) {
		max_segment = return_segment = list.length ();
		path = new PathSegment[list.length()];
	} else {
		return_segment = list.length ();
		max_segment = return_segment*2;
		path = new PathSegment[list.length()*2];
	}

	unsigned int i = 0;

	while (i < list.length ()) {
		e = list.item (i).toElement();

		// first get segment-specific animation, if available
		path[segments].z = cdp->get_int_attribute ( "z", e, 0, 300, global_z, Info );

		if (e.hasAttribute ("text") || e.hasAttribute ("file_pattern")) {
			path[segments].anim = new Animation (&e, ReadSound | ReadAnimContinuous | ReadAnimRepetitions | ReadAnimOscillating | DefaultAnimContinuous,
											sound_opts);
		} else {
			if (have_global_anim) {
				cdp->debug_msg (i18n ("Using default animation for this path-segment"), Info, &e);
				path[segments].anim = globalanim;
			} else {
				cdp->debug_msg (i18n ("Neither 'text' nor 'file-pattern'-tags were found for this segment and no default animation is available."), Critical, &e);
			}
		}

		path[segments].visible = cdp->get_bool_attribute ( "visible", e, true, Info );

		// get action-stuff (things like halt/freeze/jump)
		path[segments].action = 0;
		QDomNode n_b;
		QDomElement e_b;
		n_b = e.firstChild ();
		bool first_action = true;

		while (!n_b.isNull ()) {
			e_b = n_b.toElement ();
			QString tagName = e_b.tagName ();

			if ((tagName == "stop") || (tagName == "freeze") || (tagName == "trigger") || (tagName == "jump")) {
				Action *new_action = new Action;
				if (tagName == "stop") {
					new_action->stop = true;
					new_action->freeze = false;
					new_action->trigger = false;
					new_action->jump = false;
				} else if (tagName == "freeze") {
					new_action->stop = false;
					new_action->freeze = true;
					new_action->trigger = false;
					new_action->jump = false;
				} else if (tagName == "trigger") {
					new_action->stop = false;
					new_action->freeze = false;
					new_action->trigger = true;
					new_action->jump = false;
					QString ev_dummy = cdp->get_string_attribute ("event", e_b, bad_string_value, Serious);
					new_action->event_to_trigger = level->resolve_event (ev_dummy);
				} else if (tagName == "jump") {
					new_action->stop = false;
					new_action->freeze = false;
					new_action->trigger = false;
					new_action->jump = true;
					new_action->jump_target = 	cdp->get_int_attribute ("target", e_b, 1, max_segment, 1, Serious) -1;	// start counting from 1 instead of 0
				}
				new_action->conditional = new Conditional (cdp->get_element ("condition", e_b, Critical));
				new_action->next = 0;

				if (!first_action) {		// find place to insert the action
					Action *action_dummy = path[segments].action;
					while (action_dummy->next) {
						action_dummy = action_dummy->next;
					}
					action_dummy->next = new_action;
				} else {
					 path[segments].action = new_action;
				}
				first_action = false;
			} else {
				cdp->debug_msg (i18n ("Tag '%1' not supported. Ignored").arg(tagName), Warn, &e);
			}
        		n_b = n_b.nextSibling ();
		}

		// get initial position
// TODO: Set sensible upper and lower limits
		path[segments].start_x = cdp->get_float_attribute ( "x", e, -100, 10000, 20, Serious );
		path[segments].start_y = cdp->get_float_attribute ( "y", e, -100, 10000, 20, Serious );

		path[segments].steps = cdp->get_int_attribute ( "time", e, OBJECTS_AP, 100000, 600, Serious ) / OBJECTS_AP;

		// and now the tricky part: do path calculations
		int mode_dummy = 0;		// 0 is abrupt (easy, default), 1 is linear, 2 is custom
		dummy = cdp->get_multi_choice_attribute ( "accel", e, "linear;abrupt;custom", "abrupt", Warn );
		if ( dummy == "linear" ) {
			mode_dummy = 1;
		} else if ( dummy == "custom" ) {
			mode_dummy = 2;

			int r = (int) (1000 / OBJECTS_AP);

			// no need to know about target, Do everything right here.
			path[segments].vx = cdp->get_float_attribute ( "vx", e, -10000, 10000, 0, Warn ) / r;
			path[segments].vy = cdp->get_float_attribute ( "vy", e, -10000, 10000, 0, Warn ) / r;

			path[segments].ax = cdp->get_float_attribute ( "ax", e, -10000, 10000, 0, Warn ) / (r * r);
			path[segments].ay = cdp->get_float_attribute ( "ay", e, -10000, 10000, 0, Warn ) / (r * r);

			path[segments].dax = cdp->get_float_attribute ( "dax", e, -10000, 10000, 0, Info ) / (r * r * r);
			path[segments].day = cdp->get_float_attribute ( "day", e, -10000, 10000, 0, Info ) / (r * r * r);

		}

		float target_x, target_y;

		i++; // we need to know about the starting-points of the next frame

		if (i >= list.length ()) {
			if (looping) {
				target_x = path[0].start_x;
				target_y = path[0].start_y;
			} else {
				target_x = path[segments].start_x;
				target_y = path[segments].start_y;
			}
		} else {
			e = list.item (i).toElement ();
			target_x = cdp->get_float_attribute ( "x", e, -100, 10000, 20, Serious );
			target_y = cdp->get_float_attribute ( "y", e, -100, 10000, 20, Serious );
		}

		if ( mode_dummy == 0 ) {		// accel-mode "abrupt"
			previous_vx = path[segments].vx = (target_x - path[segments].start_x) / path[segments].steps;
			previous_vy = path[segments].vy = (target_y - path[segments].start_y) / path[segments].steps;
			path[segments].ax =	path[segments].ay = path[segments].dax = path[segments].day = 0;
		} else if ( mode_dummy == 1 ) {
			path[segments].vx = previous_vx;		// Starting with end-velocity of last segment
			path[segments].vy = previous_vy;

			float target_av_vx, target_av_vy;		// the average velocity we have to achieve
			target_av_vx = (target_x - path[segments].start_x) / path[segments].steps;
			target_av_vy = (target_y - path[segments].start_y) / path[segments].steps;
			path[segments].ax = ((target_av_vx - previous_vx) * 2) / path[segments].steps;
			path[segments].ay = ((target_av_vy - previous_vy) * 2) / path[segments].steps;

			path[segments].dax = path[segments].day = 0;

			previous_vx += path[segments].ax * path[segments].steps;
			previous_vy += path[segments].ay * path[segments].steps;
		}  /* else if ( mode_dummy == 2 ) {
		// already done
		} */

		segments++; 		// already advanced n to next element in list
	}

	if (!looping) {		// bouncing object: calculate way back
		for (int i = 0; i < segments; i++) {

			if (i < (segments -1)) {
				path[segments + i].action = path[segments-i-2].action;
			} else {
				path[segments + i].action = 0;
			}

			path[segments + i].anim = path[segments-i-1].anim;
			path[segments + i].visible = path[segments-i-1].visible;
			path[segments + i].z = path[segments-i-1].z;
			path[segments + i].steps = path[segments-i-1].steps;

			// I'd be surprised if the following section always worked as expected!

			path[segments + i].start_x = (path[segments-i-1].start_x + path[segments-i-1].steps*path[segments-i-1].vx + (path[segments-i-1].steps*path[segments-i-1].steps*path[segments-i-1].ax/2) + (path[segments-i-1].steps*path[segments-i-1].steps*path[segments-i-1].steps*path[segments-i-1].dax/6));
			path[segments + i].start_y = (path[segments-i-1].start_y + path[segments-i-1].steps*path[segments-i-1].vy + (path[segments-i-1].steps*path[segments-i-1].steps*path[segments-i-1].ay/2) + (path[segments-i-1].steps*path[segments-i-1].steps*path[segments-i-1].steps*path[segments-i-1].day/6));

			path[segments + i].vx = - (path[segments-i-1].vx + path[segments-i-1].steps*path[segments-i-1].ax + (path[segments-i-1].steps*path[segments-i-1].steps*path[segments-i-1].dax/2));
			path[segments + i].vy = - (path[segments-i-1].vy + path[segments-i-1].steps*path[segments-i-1].ay + (path[segments-i-1].steps*path[segments-i-1].steps*path[segments-i-1].day/2));

			path[segments + i].ax = (path[segments-i-1].ax + path[segments-i-1].steps*path[segments-i-1].dax);
			path[segments + i].ay = (path[segments-i-1].ay + path[segments-i-1].steps*path[segments-i-1].day);

			path[segments + i].dax = - path[segments-i-1].dax;
			path[segments + i].day = - path[segments-i-1].day;

		}
		segments *= 2;
	}
}

/** Whether the object would currently be harmful upon collision with the taxi */
bool MovingObject::harmful ()
{
	return (visible () && harmful_if_visible);
}

/** Returns the current x-velocity of the object */
float MovingObject::current_vx (){
	return (vx);
}

/** Returns the current y-Velocity of the object */
float MovingObject::current_vy(){
	return (vy);
}

void MovingObject::timerTick(int id)
{
	if (id == advance_timer) {
		if ( step == 0 ) {		// Segments has just begun
			processing_action = actions_done = false;
			move ( path[current_segment].start_x, path[current_segment].start_y );
			if (platform) platform_p->movedTo (x (), y());
			setZ (path[current_segment].z);

			// don't (re)set animation, if animation is the same as last segment!
			if ((getAnimation () != path[current_segment].anim) || (current_segment == 0)) {
				setAnimation (path[current_segment].anim, (current_segment >= return_segment) ? AnimForward : AnimReverse);
				if (animation_timer != -1) tm->killTimer (animation_timer);
				if (getAnimation ()->multiFrame ()) {
					animation_timer = tm->startTimer (this, getAnimation ()->framePeriod ());
				} else {
					animation_timer = -1;
				}
			} else {
				if (current_segment >= return_segment) {
					setDirection (AnimForward);
				} else {
					setDirection (AnimReverse);
				}
			}
			startSound (1);

			if (last_visible != path[current_segment].visible) {
				setVisible (last_visible = path[current_segment].visible);
				if (platform) platform_p->set_visible (last_visible);
				if (teleporter) teleporter_p->set_visible (last_visible);
			}
			vx = path[current_segment].vx;
			vy = path[current_segment].vy;
			ax = path[current_segment].ax;
			ay = path[current_segment].ay;
			step++;
		} else {		// normal move or action
			if (processing_action) {
				process_actions ();
				if (actions_done) {
					step = 0;
					if ( ++current_segment >= segments ) {
						current_segment = 0;
					}
					return; 		// start over with step == 0!
				}
			}
			if (!processing_action) {			// this is not an else!
				ax += path[current_segment].dax;
				ay += path[current_segment].day;
				vx += ax;
				vy += ay;
				moveBy (vx, vy);
				if (platform) platform_p->movedBy (vx, vy);
				if ( ++step >= (path[current_segment].steps -1)) { // finished this segment
					if (path[current_segment].action && (!actions_done)) {
						current_action = path[current_segment].action;
						process_actions ();
					}
					if (!processing_action) {
						step = 0;
						if ( ++current_segment >= segments ) {
							current_segment = 0;
						}
					}
				}
			}
		}
	} else if (id == animation_timer) {
		if ( tick () ) {
			tm->killTimer (animation_timer);
			animation_timer = -1;
		}
	}
}

/** Process a (chain of) actions (stop/freeze/jump) */
void MovingObject::process_actions () {
	bool processing = true;
	while (processing) {
		if (!current_action->conditional->fulfilled ()) {
			if (current_action->next) {
				current_action = current_action->next;
			} else {
				processing = false;
				processing_action = false;
				actions_done = true;
			}
		} else {
			if (initialize_action ()) {
				processing = false;
			} else {
				if (current_action->next) {
					current_action = current_action->next;
				} else {
					processing = false;
					processing_action = false;
					actions_done = true;
				}
			}
		}
	}
}

/** Does anything that needs to be done when first encountering the current_action. Returns
true, if the action is enduring until its condition fails, false, if we can proceed. */
bool MovingObject::initialize_action () {
	processing_action = true;
	if (current_action->freeze) {
		if (animation_timer != -1) {
			tm->killTimer (animation_timer);
			animation_timer = -1;
		}
		return true;
	} else if (current_action->stop) {
		return true;
	} else if (current_action->trigger) {
		current_action->event_to_trigger->trigger ();
		return false;
	} else if (current_action->jump) {
		processing_action = false;
		actions_done = true;
		current_segment = current_action->jump_target -1;		// (current_segment will be increased by one after this!)
		return true;
	} else {
		return false;
	}
}
