/***************************************************************************
                          passenger.cpp  -  description
                             -------------------
    begin                : Mon May 28 2001
    copyright            : (C) 2001, 2002 by Thomas Friedrichsmeier
    email                : Thomas.Friedrichsmeier@ruhr-uni-bochum.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "defines.h"

#include <qwidget.h>
#include <qdom.h>
#include <qstring.h>

#include <stdlib.h>

#include "passenger.h"

#include "taxipilot.h"
#include "cdp.h"
#include "platform.h"
#include "soundwrapper.h"

Passenger::Passenger(const QString & file):AnimatedSprite()
{
	idle_timer = boarding_timer = leaving_timer = animation_timer = -1;

	read_passenger(file);
	setZ (PASSENGER_Z);

	on_platform = false;
}

Passenger::~Passenger()
{
	tm->killTimers (this);
	discard ();

	delete still;
	delete teleporting;

	for (unsigned int i=0; i < num_idle_animations; i++) {
		delete idle[i];
	}
	delete[]idle;

	if (cdp->for_real () || cdp->thorough_check ()) {
		delete walking;
	}
}

/** Initializes the passenger (use on=0 for start_screen-passengers) */
void Passenger::initialize (int base_x, int base_y, Platform *on, bool for_credits) {

	setAnimation (still); // for width and heigth

	sprite_base_x = base_x - (width() / 2);
	sprite_dest_x = 0;

	if (!((on == 0) || for_credits)) {
		for_real = true;
		platform_p = on;
		on_platform = true;
		platform_p->attach_passenger (this);
	} else {
		for_real = false;
		on_platform = false;
	}

     	impatience = 1 + (rand() % 4);

	emitted_im_here = false;

	move(sprite_base_x, (base_y - height()));

	if (for_credits) {
		do_animation (still, &Passenger::do_nothing);
		idle_timer = tm->startTimer(this, (rand() % 2000) + (int) (8000 / impatience));
	} else {
		do_animation (teleporting, &Passenger::wait_for_pickup, true);
		show();
	}

}

/** Discards this passenger (but does not delete it). Call initialize in oder to reuse the
passenger */
void Passenger::discard (){

	stopSound ();
	hide ();
	if (on_platform) platform_p->detach_passenger ();
	on_platform = false;

	tm->killTimers (this);
	idle_timer = boarding_timer = leaving_timer =  animation_timer = -1;

}

/** taxi has landed at taxi_x, we can start boarding now. */
void Passenger::pick_up(int taxi_x)
{

	sprite_dest_x = taxi_x - (int) (width() / 2);
	if (leftEdge() < sprite_dest_x) {
		step = walking_speed;				// never do steps > 1! You might miss your destination.
		setOrientation (AnimRight);
	} else {
		step = -walking_speed;
		setOrientation (AnimLeft);
	}

	do_animation (walking, &Passenger::do_nothing, false, true);

	if (idle_timer != -1) {
		tm->killTimer(idle_timer);	// stop any idle animation running
		idle_timer = -1;
	}
	if (boarding_timer == -1) {	// in case we got abandoned and timer is still running
		boarding_timer = tm->startTimer(this, (int) (PASSENGER_AP / impatience));	// start boarding-animation
	}
	
	got_abandoned = false;
}

/** You got me, where I wanted to go. Thanks */
void Passenger::got_delivered(int base_x, Platform *on)
{

	platform_p = on;
	on_platform = true;
	platform_p->attach_passenger (this);

	sprite_base_x = base_x - (width() / 2);
	move(sprite_base_x, platform_p->current_plat_y() - height());
	show();

	sprite_dest_x = platform_p->current_plat_base_x ()- (width() / 2);

	if (sprite_base_x < sprite_dest_x) {
		step = walking_speed;
		setOrientation (AnimRight);
	} else {
		step = -walking_speed;
		setOrientation (AnimLeft);
	}

	// before we can leave, we have to appear, though
	do_animation (teleporting, &Passenger::leave, true);
}

void Passenger::read_passenger(const QString & file)
{								// initialization only

	QDomElement e;				// just some element
	QDomNodeList list;			// just a list of nodes
	QString dummy;

	int sound_options = DefaultSingleSound | ReadRepetitious | ReadPitchable | DefaultPitchable | ReadInterruptable;

	QDomElement xdocElement = cdp->open_config_file(file);

	author = cdp->get_string_attribute("author", xdocElement, i18n("Anonymous"), Info);
	title = cdp->get_string_attribute("title", xdocElement, i18n("Unnamed"), Info);

	e = cdp->get_element("idle", xdocElement, Critical);	// find the idle_animations-section
	list = cdp->get_node_list("data", e, 1, -1, 1, Critical);	// get all idle-animations

	num_idle_animations = 0;
	idle = new Animation * [list.length()];

	while (num_idle_animations < list.length ()) {
		e = list.item (num_idle_animations).toElement();

		idle[num_idle_animations] = new Animation (&e, ReadAnimOscillating | ReadAnimRepetitions | ReadSound, sound_options | ReadContinuous);

		num_idle_animations++;
	}

	e = cdp->get_element("teleporting", xdocElement, Critical);	// find the teleporting-animation
	teleporting = new Animation (&e, ReadSound, 0);

	e = cdp->get_element("still", xdocElement, Critical);	// find the standing-still-image
	// not sure, we'll ever use different frames on this, though
	still = new Animation (&e, ReadAnimOscillating | DefaultAnimContinuous | ReadSound, sound_options | DefaultInterruptable | ReadContinuous | DefaultContinuous);

	// find the global sounds
	sound_hello.create (cdp->get_element("hello", xdocElement, Info), DefaultSingleSound | ReadPitchable | DefaultPitchable, Info);

	// the following are only needed for the actual game
	if (cdp->for_real() || cdp->thorough_check ()) {
		sound_goodbye.create (cdp->get_element("goodbye", xdocElement, Info), DefaultSingleSound | ReadPitchable | DefaultPitchable, Info);
		sound_die.create (cdp->get_element("die", xdocElement, Info), DefaultSingleSound | ReadPitchable | DefaultPitchable, Info);
		sound_abandoned.create (cdp->get_element("abandoned", xdocElement, Info), DefaultSingleSound | ReadPitchable | DefaultPitchable, Info);

		e = cdp->get_element("walking", xdocElement, Critical);
		walking_speed = cdp->get_float_attribute ("speed", e, 0.05, 5, 0.6, Info);

		walking = new Animation (&e, ReadAnimOscillating | DefaultAnimContinuous | ReadSound, sound_options | DefaultInterruptable | ReadContinuous | DefaultContinuous, "_left", "_right");
	}

}

/** Moves the passenger and the reference points by the given offset-values. Used
by moving platforms to move an attached passenger. */
void Passenger::move_passenger_by (float delta_x, float delta_y){

	sprite_base_x += delta_x;
	sprite_dest_x += delta_x;
	// passenger does not change y-position on its own and therefore does never store it

	moveBy (delta_x, delta_y);

}

/** Sets of the animation anim for this passenger,  and trigger
a function-call when completed. Reverse means: play in reverse order; dep_on_impatience means animation
speed depends on passenger impatience */
void Passenger::do_animation (Animation * anim, void (Passenger::*call_back) (), bool reverse, bool dep_on_impatience) {

	this->next_call_back = call_back;
	if (reverse) {
		setAnimation (anim, AnimReverse);
	} else {
		setAnimation (anim, AnimForward);
	}
	if (animation_timer != -1) {
		tm->killTimer (animation_timer);
	}	
	if (!dep_on_impatience) {
		animation_timer = tm->startTimer (this, anim->framePeriod ());
		startSound (1);
	} else {
		animation_timer = tm->startTimer (this, anim->framePeriod () / impatience);		
		startSound (impatience);
	}
}

/** Function called when the passenger has just appeared on the platform, or
when it had been abandoned and is back to its base-position. */
void Passenger::wait_for_pickup (){

	do_animation (still, &Passenger::do_nothing);         // this is a continuous animation. no call_back needed
	idle_timer = tm->startTimer(this, (rand() % 2000) + (int) (8000 / impatience));
	if (!emitted_im_here) {
		if (for_real) {
			tp->passenger_ready ();
			sw->start_sound (sound_hello.id[0], pitch (sound_hello.properties & Pitchable));
		}
		emitted_im_here = true;
	}

}

/** Function called when the passenger has been delivered (and is done teleporting
from the taxi) and is to leave the platform. */
void Passenger::leave (){

	sw->start_sound (sound_goodbye.id[0], pitch (sound_goodbye.properties & Pitchable));
	do_animation (walking, &Passenger::do_nothing, false, true);
	leaving_timer = tm->startTimer(this, (int) (PASSENGER_AP / impatience));

}

/** Function called when the passenger has disappeared completely. */
void Passenger::done (){

	hide ();
	tp->delivery_complete ();
	// passenger will subsequently get discarded -> no call to platform_p->detach ();
	
}

/** Function called when the passenger is ready teleporting into the taxi. */
void Passenger::boarded (){

	hide();
	platform_p->detach_passenger();
	on_platform = false;
	tp->picked_up ();

}

/** Does nothing */
void Passenger::do_nothing (){
}

void Passenger::timerTick (int id) {

	if (id == boarding_timer) {
		moveBy(step, 0);
		if (tp->onActivePlatform()) {
			if ((leftEdge() - sprite_dest_x) * step >= 0) {

				tm->killTimer(boarding_timer);
				boarding_timer = -1;

				// teleport to vanish inside taxi
				do_animation (teleporting, &Passenger::boarded);
			}
		} else {			// Hey, where did you go?!
			if (!got_abandoned) {	// we've only just been abandoned
				sprite_dest_x = sprite_base_x;	// head back for base
				step = (step * -1);

				if (step > 0) {                         	// turn around
					setOrientation (AnimRight);
				} else {
					setOrientation (AnimLeft);
				}

				do_animation (walking, &Passenger::do_nothing, false, true);

				sw->start_sound (sound_abandoned.id[0], pitch (sound_abandoned.properties & Pitchable));
	
				got_abandoned = true;
			} else {		// we've long since been abandoned
				moveBy(step, 0);
				if ((leftEdge() - sprite_dest_x) * step >= 0) {
					tm->killTimer(boarding_timer);	// back to square 1
					boarding_timer = -1;

					wait_for_pickup ();
				}
			}
		}
	} else if (id == leaving_timer) {
		if ((leftEdge() - sprite_dest_x) * step >= 0) {
			tm->killTimer(leaving_timer);
			leaving_timer = -1;

			// teleport to vanish
			do_animation (teleporting, &Passenger::done);
		} else {
			moveBy(step, 0);
		}
	} else if (id == idle_timer) {	// this only sets off the actual animation
		tm->killTimer (idle_timer);
		do_animation (idle[(rand() % num_idle_animations)], &Passenger::wait_for_pickup, false, true);
	} else if (id == animation_timer) {
		if (tick ()) {
			tm->killTimer (animation_timer);
			animation_timer = -1;
			(this->*next_call_back) ();
		}
	}
}

/** Adjusts the passenger to a fractional position, so it doesn't wiggle on a moving platform */
void Passenger::adjust(float fraction_x, float fraction_y){
	float old_x = x();
	move (((int) x()) + fraction_x, ((int) y()) + fraction_y);
	sprite_dest_x += x () - old_x;
	sprite_base_x += x () - old_x;
}

/** just a really primitve abbreviation */
float Passenger::pitch (bool pitchable){
	if (pitchable) {
		return (float) impatience;
	} else {
		return (float) 1;
	}
}

/** Tells the passenger to die (it's a wild world!), i.e. make an appropriate sound(/anim?) if any. */
void Passenger::die (){
	// just in case there are still other sounds playing
	stopSound ();

	sw->start_sound (sound_die.id[0], pitch (sound_die.properties & Pitchable));
}
