/***************************************************************************
                          soundhandle.cpp  -  description
                             -------------------
    begin                : Mon Aug 5 2002
    copyright            : (C) 2002 by Thomas Friedrichsmeier
    email                : tfry@users.sourceforge.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "soundhandle.h"

#include <qdom.h>
#include <qstring.h>
#include "klocale.h"
#include "cdp.h"
#include "soundwrapper.h"

SoundHandle::SoundHandle(){
	id[0] = id[1] = 0;
	properties = 0;
}
SoundHandle::~SoundHandle(){
	sw->unregister_sound (id[0]);	// if sounds have never been registered, ids are zero and Soundwrapper will ignore this.
	sw->unregister_sound (id[1]);
}

/** Reads the information for this sound from the XML-file. See enum TPBase::SoundConfigOptions for options-parameters.
	Level specifies debug-level if no sound-tag is found. For bad options etc. debug-level Warn will be used */
void SoundHandle::create (QDomElement const & parent, const int options, DebugLevel level) {
	QDomElement e;	
	QString dummy;   			// saves us writing a few constant strings
	QString file, filea, fileb;

	e = cdp->get_element ("sound", parent, level);	

	if (!e.isNull ()) {		// Skip if there is no sound-tag

		if (options & ReadPitchable) {
			if (e.hasAttribute (dummy = "pitchable")) {
				if (cdp->get_bool_attribute (dummy, e, options & DefaultPitchable, Warn)) {
					properties |= Pitchable;
				}
			}
			else {
				if (options & DefaultPitchable) {
					properties |= Pitchable;
				}
			}
		}

		if (options & ReadContinuous) {
			if (e.hasAttribute (dummy = "continuous")) {
				if (cdp->get_bool_attribute (dummy, e, options & DefaultContinuous, Warn)) {
					properties |= Continuous;
				}
			}
			else {
				if (options & DefaultContinuous) {
					properties |= Continuous;
				}
			}
		}

		if (options & ReadRepetitious) {
			if (e.hasAttribute (dummy = "repetitious")) {
				if (cdp->get_bool_attribute (dummy, e, options & DefaultRepetitious, Warn)) {
					properties |= Repetitious;
				}
			}
			else {
				if (options & DefaultRepetitious) {
					properties |= Repetitious;
				}
			}
		}

		if (options & ReadInterruptable) {
			if (e.hasAttribute (dummy = "interruptable")) {
				if (cdp->get_bool_attribute (dummy, e, options & DefaultInterruptable, Warn)) {
					properties |= Interruptable;
				}
			}
			else {
				if (options & DefaultInterruptable) {
					properties |= Interruptable;
				}
			}
		}

	// The following code looks more complicated than is what it does. Most is just for debugging purposes.
		bool looks_like_two_files = false;
		bool is_two_files = false;
		bool has_single_file = false;
		bool use_single = true;

		fileb = "file_b";		// wouldn't get there in the next line, if a file_a-attribute is given
		if (e.hasAttribute (filea = "file_a") || e.hasAttribute (fileb)) {
			looks_like_two_files = true;		
			if (e.hasAttribute (filea) && e.hasAttribute (fileb)) {
				is_two_files = true;
			}
		}
		if (e.hasAttribute (file = "file")) {
			has_single_file = true;
		}

		if (has_single_file && looks_like_two_files) {
			cdp->debug_msg (i18n ("Both a 'file'- and at least one of 'file_a' or 'file_b'-attributes have been specified. Only either is allowed"), Warn, &e);
		}
		if (looks_like_two_files && (!is_two_files)) {
			cdp->debug_msg (i18n ("A 'file_a'-attribute has been given without a 'file_b'-attribute (or vice versa). Using only a single file"), Warn, &e);			
		}
		if (!(has_single_file || looks_like_two_files)) {
			cdp->debug_msg (i18n ("No file specified. Sound will not be played."), Warn, &e);			
			return;
		}

		dummy = bad_string_value;
		if (options & ReadSingleSound) {			
			// in this case, DefaultSingleSound is ignored, since if would not make any sense
			if (is_two_files) {
				filea = cdp->get_string_attribute (filea, e, dummy, Warn);
				fileb = cdp->get_string_attribute (fileb, e, dummy, Warn);
				use_single = false;
			} else {
				if (has_single_file) {
					file = cdp->get_string_attribute (file, e, dummy, Warn);				
				} else {
					if (e.hasAttribute (filea)) {
						file = cdp->get_string_attribute (filea, e, dummy, Warn);				
					} else {		// has "file_b"
						file = cdp->get_string_attribute (fileb, e, dummy, Warn);				
             			}
				}
			}
		} else {
			if (options & DefaultSingleSound) {
				if (is_two_files) {
					cdp->debug_msg (i18n ("Two sounds have been given, where only one can be used."), Warn, &e);			
				}
				if (has_single_file) {
					file = cdp->get_string_attribute (file, e, dummy, Warn);				
				} else {
					if (e.hasAttribute (filea)) {
						file = cdp->get_string_attribute (filea, e, dummy, Warn);				
					} else if (e.hasAttribute (fileb)) {
						file = cdp->get_string_attribute (fileb, e, dummy, Warn);				
             			}
				}
			} else {
				if (is_two_files) {
					filea = cdp->get_string_attribute (filea, e, dummy, Warn);
					fileb = cdp->get_string_attribute (fileb, e, dummy, Warn);
					use_single = false;
				} else {
					cdp->debug_msg (i18n ("Only one sound was specified, where two were expected"), Warn, &e);							
					if (has_single_file) {
						file = cdp->get_string_attribute (file, e, dummy, Warn);				
					} else {
						if (e.hasAttribute (filea)) {
							file = cdp->get_string_attribute (filea, e, dummy, Warn);				
						} else {		// has "file_b"
							file = cdp->get_string_attribute (fileb, e, dummy, Warn);				
	      	       			}
					}
				}
			}
		}

	// Now, finally, register those sound(s)		            		
		if (use_single) {
			id[0] = id[1] = sw->register_sound (cdp->prefix_check (file, Warn), properties & Continuous, properties & Interruptable);
		} else {
			id[0] = sw->register_sound (cdp->prefix_check (filea, Warn), properties & Continuous, properties & Interruptable);
			id[1] = sw->register_sound (cdp->prefix_check (fileb, Warn), properties & Continuous, properties & Interruptable);
		}
	}
}
